;;; To compile .java files you need:

;;; src-base the root directory of your .java sources.
;;; class-base the root directory where you want .class files to go.

;;; (compile-class class-base src-base "package.Class") - compile's a class.

;;; (recompile class-base src-base package recurse?) - recompile all
;;; the files in a package, decending into subpackages if recurse? is true.

;;; There are also (javac), (jar), and (rmic).

(load "build/bootstrap.silk")
(load "elf/util.scm")

(import "java.io.File")
(import "java.io.FileInputStream")
(import "java.io.FileOutputStream")
(import "java.lang.Double")
(import "java.lang.String")
(import "java.lang.System")
;;; (import "silk.SchemeClassLoader")

;;; KRA 02JUN00: These will probably move.
(define (array a-class . args)
  (let ((v (make-array a-class (length args))))
    (let loop ((i 0)
	       (as args))
      (if (null? as) v
	  (begin
	    (vector-set! v i (car as))
	    (loop (+ i 1) (cdr as)))))))

(define (make-array a-class size)
  (java.lang.reflect.Array.newInstance a-class size))

(define-method (delete-class-files (directory java.io.File))
  (file-walk directory 
	     (lambda (f result) (if (class-file? f) (.delete f)))
	     ()))

(define (file-separator) File.separator$)

(define (separate by items)
  ;; (separate "," '(1 2 3)) -> (1 "," 2 "," 3)
  (define (separate0 head tail)
    (if (null? tail) (list head)
	(cons head (cons by (separate0 (car tail) (cdr tail))))))
  (if (null? items) items
      (separate0 (car items) (cdr items))))

(.equals (separate "," '(1 2 3)) '(1 "," 2 "," 3))

(define (class->file name)
  ;; (class->file "foo.bar.baz") -> "foo/bar/baz.java".
  (apply string-append
	 (append (separate (file-separator)
			   (crack name "." 0))
		 (list ".java"))))

(.equals (File. (class->file "foo.bar.baz")) (File. "foo/bar/baz.java"))

(define (package->file base package)
  (string-append (File. base (apply string-append
				   (separate (file-separator)
					     (crack package "." 0))))))
(.equals (package->file (File. "base/base/") "com.bbn.icis.dm")
	 "base/base/com/bbn/icis/dm")

(define (file->class name)
  ;; (file->class (File. "foo/bar.java")) -> "foo.bar"
  (let ((name (.toString name)))
    (or (and (.endsWith name ".java")
	   (apply string-append
		  (separate "."
			    (crack (substring name 0 
					      (- (string-length name) 
						 (string-length ".java")))
				   (file-separator)
				   0))))
      (error name "doesn't end with \".java\""))))

(.equals (file->class (File. "foo/bar.java")) "foo.bar")

(define (string-array args)
  (apply array (cons String.class (map .toString args))))

(define (javac . args)
  (.compile
   (sun.tools.javac.Main. System.out$ "silk-javac")
   (string-array args)))

(define (jar . args)
  (sun.tools.jar.Main.run
   (sun.tools.jar.Main. System.out$ System.err$ "silk-jar")
   (string-array args)))

(define (rmic . args)
  (.compile (sun.rmi.rmic.Main. System.out$ "silk-rmic")
	    (string-array args)))

(define (remote? c)
  ;; is c a class that needs to be rmic'd?
  ;; It is if it or one of its interfaces is java.rmi.Remote.
  (define (any p xs)
    ;; Does any element of xs satisfy (p x)?
    (and (not (null? xs)) (or (p (car xs)) (any p (cdr xs)))))
  (define (remote0? c)
    (or (eq? c java.rmi.Remote.class)
	(any remote0? (vector->list (.getInterfaces c)))))
  (and (not (.isInterface c)) (remote0? c)))

(define (compile-class class-base src-base name)
  (and (compile-file class-base src-base (class->file name))
       (let ((c (class name)))
	 ;; Sometimes a java file can have the wrong package defined
	 (if (eq? c #null) (print (list "Can't find class" name))
	     (if (remote? c)
		 (begin
		   (display "rmicing ") (display name) (newline)
		   (rmic "-classpath" (System.getProperty "java.class.path")
			 "-d" class-base
			 ;; "-sourcepath" src-base
			 name))))
	 #t)))

(define (compile-file class-base src-base file)
  ;; Compile the *.java file, file using the current CLASSPATH.
  (let* ((separator (System.getProperty "path.separator")))
    (javac
     "-g" "-deprecation" "-classpath"
     (string-append (System.getProperty "java.class.path") separator
		    class-base)
     "-d" class-base
     "-sourcepath" src-base
     (.toString (make-src-file src-base file)))))

'(define class-base (File. "d:/temp"))
'(define src-base (File. "d:/silk4/silkweb/"))
'(compile-file class-base src-base "silk/Invoke.java")
'(compile-class class-base src-base "silk.Invoke")

(define (file->package base file)
  (let* ((base (.toString base))
	 (L (string-length base))
	 (package (separate
		   "."
		   (crack (.substring (.toString file)
				      (if (.endsWith base (file-separator))
					  L
					  (+ L 1)))
			  (file-separator)
			  0))))
    (apply string-append package)))

(.equals (file->package (File. "base/") (File. "base/bar/baz"))
	 (file->package (File. "base") (File. "base/bar/baz")))

(.equals (file->package (File. "base/") (File. "base/bar/baz")) "bar.baz")

(define (relativize-file base file)
  ;; (relativize-file "\\frog\\prince\\" "\\frog\\prince\\bar.java")
  ;; -> bar.java.
  (let* ((string (.toString base))
	 (L (.length string)))
    (File. (.substring (.toString file)
		       (if (.endsWith string (file-separator)) L (+ L 1))))))

(define (make-class-file class-base java-file)
  (let ((name (.getName java-file)))
    (File. class-base 
	   (string-append
	    (.getParent java-file)
	    File.separator$
	    (substring name 0 (- (string-length name) 
				 (string-length ".java")))
	    ".class"))))

(define (make-src-file src-base java-file)
  (File. src-base (.toString java-file)))

(define (needs-recompile class-file java-file)
  (if (not (.exists java-file)) (error java-file " does not exist!"))
  (or (not (.exists class-file))
      (> (.lastModified java-file) (.lastModified class-file))))

(define (recompile class-base src-base package recurse?)
  ;; Recompile an entire package, and subpackages if recurse? is #t.
  (display "Recompiling package ") (display package)
  (if recurse? (display " recursively"))
  (newline)
  ;; Make the class directory if necessary.
  (.mkdirs (File. (package->file class-base package)))
  (let ((directory (File. (package->file src-base package))))
    (recompile-files class-base src-base (java-files directory))
    (if recurse?
	(map* (lambda (f)
		(recompile class-base src-base 
			   (file->package src-base f)
			   recurse?))
	      (directories directory)))))

(define (recompile-files class-base src-base jfs)
  (if (null? jfs) #t
      (let* ((it (relativize-file src-base (first jfs)))
	     (jfs (cdr jfs))
	     (java-file (make-src-file src-base it))
	     (bin-file (make-class-file class-base it)))
	;; A *.java file needs to be recompiled if it is younger than
	;; its *.class file or if it's class is (remote?) so rmic can be
	;; run.  

	;; +++ We could test if the stub and skeleton files are upto
	;; date.
	(if (or (needs-recompile bin-file java-file)
		(let ((c (class (file->class it))))
		  (and c (remote? c))))
	    (begin
	      (display "Compiling ")
	      (display java-file) (newline)
	      (compile-class class-base src-base 
			     (file->class it))))
	(recompile-files class-base src-base jfs))))

(define (null->empty x) (if (eq? #null x) '() x))

(define (directories directory)
  (filter-in .isDirectory (null->empty (.listFiles directory))))

(define (java-files directory)
  (filter-in java-file? (null->empty (.listFiles directory))))

(define (recompile-class class-base src-base name)
  (let* ((it (File. (class->file name)))
	 (java-file (make-src-file src-base it))
	 (bin-file (make-class-file class-base it)))
    ;; A *.java file needs to be recompiled if it is younger than
    ;; its *.class file or if it's class is (remote?) so rmic can be
    ;; run.  

    ;; +++ We could test if the stub and skeleton files are upto
    ;; date.
    (if (or (needs-recompile bin-file java-file)
	    (let ((c (class name)))
	      (and c (remote? c))))
	(begin
	  (display "Compiling ")
	  (display java-file) (newline)
	  (compile-class class-base src-base 
			 (file->class (.toString it)))))
    bin-file))

;;; KRA 23SEP99: The prototype of build.CompilingClassLoader.
'(define (compilingClassLoader class-base src-base parent)
  (SchemeClassLoader
   parent
   (lambda (name)
     (SchemeClassLoader.toBytes
      (recompile-class class-base
		       src-base
		       (symbol->string (toString name)))))))

(define-method (copyBytes (in java.lang.Object) (out java.lang.Object))
  ;; Provide default N = 1001.
  (copyBytes in out 1001))

(define-method (copyBytes (in java.io.InputStream) 
			  (out java.io.OutputStream) 
			  N)
  (let ((bs (make-array (class "byte") N)))
    (let loop ((i (.read in bs 0 N)))
      (if (not (= i -1))
	  (begin
	    (.write out bs 0 i)
	    (loop (.read in bs 0 N)))))))

(define-method (copyBytes (in java.io.File) (out java.io.File) N)
  (let ((in (FileInputStream. in))
	(out (FileOutputStream. out)))
    (copyBytes in out N)
    (.close in)
    (.close out)))

(define-method (copyBytes (in java.net.URL) (out java.io.File) N)
  (let ((in (.openStream in))
	(out (FileOutputStream. out)))
    (copyBytes in out N)
    (.close in)
    (.close out)))

(define (do-files from-directory to-directory predicate act)
  (assert (and (.isDirectory from-directory) (.isDirectory to-directory)))
  (file-walk from-directory
	     (lambda (f so-far) 
	       (if (predicate f) 
		   (act f (swing-file from-directory to-directory f))))
	     '()))

(define (swing-file from to file)
  (let ((sfile (.toString file))
	(sfrom (.toString from)))
    (assert (= (.indexOf sfile sfrom) 0))
       (let ((relative (.substring sfile (+ (.length sfrom) 1))))
	 (File. to relative))))
  
(.equals (swing-file (File. "d:/silk4/silkweb/silk")
		     (File. "d:/temp/silk")
		     (File. "d:/silk4/silkweb/silk/Invoke.java"))
	 (File. "d:/temp/silk/Invoke.java"))

(define (copy-files from to predicate)
  (assert (not (eq? from to)))
  (.mkdirs to)
  (do-files from to predicate 
	    (lambda (f1 f2)
	      ;; (print `(copy ,f1 ,f2))
	      (mkdirs (getParentFile f2))
	      (copyBytes f1 f2))))

;;; com.sun.tools.javadoc.Main.main